%shorttrace tom
/********************************************************************/
/*                                                                  */
/*             Turbo Prolog 2.0 example program                     */
/*								    */
/*   Copyright (c) 1986, 88 by Borland International, Inc.          */
/*                                                                  */
/*			 SENTENCE ANALYSIS			    */
/* This sample shows how sentence analysing can be done in TURBO    */
/* PROLOG.							    */
/* The structure of sentences is nicely modelled in the typesystem, */
/* and the resulting parse-tree illustrates how easily these things */
/* can be doen in Turbo Prolog					    */
/*								    */
/* As an example the following sentence can be recognized:	    */
/*  - every man that lives loves a woman			    */
/*								    */
/********************************************************************/
/*
DATABASE   
  * words which can be recognized *
  det( STRING )
  noun( STRING )
  rel( STRING )
  verb( STRING )

DOMAINS
  DETERM   = none ; determ( STRING )
  NOUNP    = nounp( DETERM, STRING, RELCL)
  RELCL    = none ; relcl( STRING, VERBP )
  SENTENCE = sent( NOUNP, VERBP )
  TOKL     = STRING*
  VERBP    =	verb( STRING ) ; verbp( STRING, NOUNP );
  		verbc(STRING, DETERM)	% is heavy
  FILE	= out; err; tree_out

* tree definition *
  CONTENTS=ITEM*
  ITEM=c(STRING,TTYPE)
  TTYPE=t(TREE); text(STRING); pict(STRING)
  TREE = tree(STRING,CONTENTS,SELECTOR,TREELIST); none
  TREELIST = TREE*

*/
include "maktree.pro"

PREDICATES
/*
  Recognition of words in different forms
*/
  is_det( STRING )
  is_noun( STRING )
  is_rel( STRING )
  is_verb( STRING )

/*
  Parser
*/
  s_determ(   TOKL, TOKL, DETERM )
  s_nounp(    TOKL, TOKL, NOUNP )
  s_relcl(    TOKL, TOKL, RELCL )
  s_sentence( TOKL, TOKL, SENTENCE )
  s_verbp(    TOKL, TOKL, VERBP )
  write_tok(TOKL)
  add_word(INTEGER,STRING)
  is_complex_verb(TOKL)
  is_complex(STRING)
  is_end_symbol(STRING)

/*
  scanner
*/
  check(STRING)
  tokl( COL, COLL, STRING, TOKL )
  tom(TOKL,TOKL)

/*
  Main predicates
*/
  analyze
  key
%  process(INTEGER)
  run1
  run2(TOKL, COLL)
%  sen_an
  execute_branch(INTEGER)

/*
  Update database predicates
*/
  read(STRING,STRING)
%  updatdba
  updatdba1(INTEGER)

CLAUSES

key:-
	write("\n\n>> Press any key: "),
	readkey(_).

execute_branch(1001):-/*load text*/
		attention("load text",""),!.
execute_branch(1002):-/*load database*/
		retractall(_,sen_base),
		consult("sen_an.san",sen_base),!;
		attention("Can't read\n sen_an.san\n",""),!.
execute_branch(1003):-/*clear database*/
		retractall(_),!.
execute_branch(2001):-/*save text*/
		attention("save text",""),!.
execute_branch(2002):-/*save database*/
		existfile("sen_an.bak"),!,
		deletefile("sen_an.bak"),
		renamefile("sen_an.san","sen_an.bak"),
		save("sen_an.san",sen_base);
		renamefile("sen_an.san","sen_an.bak"),
		save("sen_an.san",sen_base),!.
execute_branch(2003):-/*save tree*/
		treebase(_),
		writedevice(tree_out),
		findall(X,treebase(X),TREELIST),
		write("treebase(tree(\"TEXT\",[],0,",TREELIST,"))"),
		writedevice(screen),!.
execute_branch(3001):-/*edit text*/
		attention("edit text",""),!.
execute_branch(3002):-/*edit database*/
		file_str("sen_an.san",DB),
		edit(DB,DBNEW),
		clearwindow,
		not(DB = DBNEW),
		attention("Do you wish","to save the changes?"),
		file_str("sen_an.san",DBNEW),
		retractall(_),
		execute_branch(1002),!./*load database*/
execute_branch(4001):-/*Show verbs*/
		updatdba1(1),!.
execute_branch(4002):-/*Show nounns*/
		updatdba1(2),!.
execute_branch(4003):-/*Show relatives*/
		updatdba1(3),!.
execute_branch(4004):-/*Show determs*/
		updatdba1(4),!.
execute_branch(4005):-/*Show tree*/
		show_tree,!.
execute_branch(4006):-/*Show errors*/
		closefile(err),
		file_str("error.sen",STR),
		openappend(err,"error.sen"),
		display(STR),!.
execute_branch(5001):-/*add verbs*/
		enter("Enter verb: ","",X),
		add_word(1,X),!.
execute_branch(5002):-/*add nouns*/
		enter("Enter noun: ","",X),
		add_word(2,X),!.
execute_branch(5003):-/*add relatives*/
		enter("Enter relatives: ","",X),
		add_word(3,X),!.
execute_branch(5004):-/*add determs*/
		enter("Enter determs: ","",X),
		add_word(4,X),!.
execute_branch(6001):-/*analyse sentence*/
		analyze,!.
%		attention("analyse sentence",""),!.
execute_branch(6002):-/*analyse text*/
		edit("",TEXT),
		clearwindow,
		tokl(5,POSL,TEXT,TOKL),
		run2(TOKL,POSL),
		write("Press any key.."),readchar(_),
		show_tree,!.
execute_branch(8001):-/*help analyser*/
		file_str("sen_an.hlp",TXT),
		display(TXT),
		clearwindow,!;
		attention("Can't read sen_an.hlp",""),!.
execute_branch(8002):-/*help treebrowser*/
		attention("help treebrowser",""),!.
execute_branch(-1):-attention("Exit program\nDo you want it?","y/n"),exit,!.
execute_branch(_).
/* * * * * * * * * * *
  Analyze a sentence
* * * * * * * * * * */  
analyze:-
	shiftwindow(screen_window),
    clearwindow, 
    write("Try: every man that lives loves a woman\n"),
    run1,!;
    clearwindow.

run1:-
    write("\n Write a sentence:\n    "),
    readln(LIN), LIN >< "" ,
%    retractall(treebase(_)),
    tokl(5,POSL,LIN,TOKL),
    run2(TOKL,POSL), !,
    run1.

run2([],_):-!.
run2(TOKL,_):-
    s_sentence( TOKL, NEXTTOKL, SENT ),!,
%    cursor(21,1),
    writedevice(out),
    write_tok(TOKL),write(".\n"),
    writedevice(screen),
	make_tree(SENT,TREE),
	assert(treebase(TREE)),
    run2(NEXTTOKL,[]).
run2(_,_).

show_tree:-   
   textmode(MAXROW,_),NROW=MAXROW-1,
   makewindow(tree_window,111,23,"",0,0,NROW,65),
   makestatus(112,status_line),
   newupdate(treewindow(0,0)),
   treebase(_),
   findall(X,treebase(X),TREELIST),
   repeat,   
   cursor(RC,CC),
   clearwindow,
   cursor(RC,CC),
   treemenu(right,tree("TEXT",[],0,TREELIST),_,KEY),
   KEY=esc,!,
   removestatus,
   removewindow(tree_window,1).

tokl(POS,[POS1|POSL],STR,[TOK|TOKL]) :-
    fronttoken(STR,TOK,STR1),
    check(TOK),!,
    str_len(TOK,LEN),
    POS1=POS+(LEN+1) div 2,
    POS2=POS+5+LEN,
    tokl(POS2,POSL,STR1,TOKL).
tokl(_,[],_,[]).

s_sentence(TOKL,OUTTOKL,sent(NOUNP,VERBP)):-
	s_nounp(TOKL,TOKL1,NOUNP),
	s_verbp(TOKL1,TOKL2,VERBP),
	tom(TOKL2,OUTTOKL),!.
s_sentence(TOKL,_,_):-
	writedevice(err), write_tok(TOKL),nl,
	writedevice(screen),
	write_tok(TOKL),nl,
	write(">> Sentence not recognized.\n"),fail.

tom([],[]):-!.
tom([H|TOKL],TOKL):-is_end_symbol(H).

write_tok([H|_]):-is_end_symbol(H),!.
write_tok([H|T]):-write(H," "),!,
	write_tok(T).
write_tok(_).

s_nounp(TOKL,TOKL2,nounp(DETERM,NOUN,RELCL)):-
	s_determ(TOKL,[NOUN|TOKL1],DETERM),
	is_noun(NOUN),
	s_relcl(TOKL1,TOKL2,RELCL),!.
/* determiner exists, noun absent - add empty noun (is heavy)*/
s_nounp(TOKL,TOKL2,nounp(DETERM,END,none)):-
	s_determ(TOKL,[END|TOKL2],DETERM),
	is_end_symbol(END),!.

s_determ([DETERM|TOKL],TOKL,determ(DETERM)):-
	is_det(DETERM),!.
s_determ(TOKL,TOKL,none).

s_relcl([REL|TOKL],TOKL1,relcl(REL,VERBP)):-
    is_rel(REL),
    s_verbp(TOKL,TOKL1,VERBP),!.
s_relcl(TOKL,TOKL,none).

s_verbp([VERB,DETERM|TOKL2],TOKL2,verbp(VERB,nounp(determ(DETERM),"",none))):-
	is_complex_verb([VERB,DETERM|TOKL2]),!.
s_verbp([VERB|TOKL],TOKL1,verbp(VERB,NOUNP)):-
	is_verb(VERB),
	s_nounp(TOKL,TOKL1,NOUNP),!.
s_verbp([VERB|TOKL],TOKL,verb(VERB)):-
	is_verb(VERB).

is_complex_verb([VERB,DETERM|_]):-
    	is_complex(VERB),
	is_det(DETERM),
	DETERM <> "a",
	DETERM <> "an",
	DETERM <> "the",!.
is_complex_verb([VERB,VERB1|_]):-
    	is_complex(VERB),
	is_verb(VERB1),!.

check(WORD):-is_noun(WORD),!.
check(WORD):-is_det(WORD),!.
check(WORD):-is_rel(WORD),!.
check(WORD):-is_verb(WORD),!.
check(WORD):-is_end_symbol(WORD),!.
check(WORD):- write(">> Unknown word: ",WORD),nl,
	longmenu(10,10,40,6,7,7,
	     ["Verb", 
	      "Noun",
	      "Relative", 
	      "Determiner"],"Choose word type",
  		1,CHOICE,_,_),
	    add_word(CHOICE,WORD),!.

add_word(1,X):-concat(Y,"s",X),!,
		assert(verb(Y)).
add_word(1,X):-concat(Y,"ed",X),!,
		assert(verb(Y)).
add_word(1,X):-concat(Y,"es",X),!,
		assert(verb(Y)).
add_word(1,X):-concat(Y,"ing",X),!,
		assert(verb(Y)).
add_word(1,X):-assert(verb(X)).
add_word(2,X):-assert(noun(X)).
add_word(3,X):-assert(rel(X)).
add_word(4,X):-assert(det(X)).

is_end_symbol(".").
is_end_symbol(",").
is_end_symbol("!").
is_end_symbol("?").

is_complex("is").
is_complex("was").
is_complex("are").
is_complex("were").

is_noun(X):-noun(X),!.
is_noun(X):-noun(Y),concat(Y,"s",X),!.

is_det(X):-det(X),!.

is_rel(X):-rel(X),!.

is_verb(X):-verb(X),!.
is_verb(X):-verb(Y),concat(Y,"s",X),!.
is_verb(X):-verb(Y),concat(Y,"ed",X),!.
is_verb(X):-verb(Y),concat(Y,"es",X),!.
is_verb(X):-verb(Y),concat(Y,"ing",X),!.

/* * * * * * * * * * * *
  Update/Show database
* * * * * * * * * * * */
/*
updatdba:-
    repeat,
    longmenu(10,20,7,7,
     ["Show verbs", 
      "Show nouns",
      "Show relatives", 
      "Show determiners",
      "",
      "Add a verb", 
      "Add a noun",
      "Add a relative", 
      "Add a determiner"],"Update/Show database",
      5,CHOICE,_),
    updatdba1(CHOICE),
    CHOICE=0,!.
*/
updatdba1(0).
updatdba1(1):-
    write("\n\nVerbs:\n******\n"),
    verb(X), write(X,' '), fail.
updatdba1(1):-nl,key,clearwindow.
updatdba1(2):-
    write("\n\nNouns:\n******\n"),
    noun(X), write(X,' '), fail.
updatdba1(2):-nl,key,clearwindow.
updatdba1(3):-
    write("\n\nRelatives:\n**********\n"),
    rel(X), write(X,' '), fail.
updatdba1(3):-nl,key,clearwindow.
updatdba1(4):-
    write("\n\nDeterminers:\n************\n"),
    det(X), write(X,' '), fail.
updatdba1(4):-nl,key,clearwindow.
updatdba1(6):-
    read("New verb",X),
    assert(verb(X)).
updatdba1(7):-
    read("New noun",X),
    assert(noun(X)).
updatdba1(8):-
    read("New relative",X),
    assert(rel(X)).
updatdba1(9):-
    read("New determiner",X),
    assert(det(X)).

read(TXT,ANS):-nl,
    write(TXT,": "),
    readln(ANS),clearwindow,ANS><"".
